import { Clear, DragHandle } from '@mui/icons-material'
import {
  Box,
  Button,
  FormControl,
  IconButton,
  InputLabel,
  MenuItem,
  Paper,
  Select,
  Table,
  TableBody,
  TableCell,
  TableContainer,
  TableHead,
  TableRow,
  TextField,
  Tooltip,
  Typography
} from '@mui/material'
import JSZip from 'jszip'
import { MuiFileInput } from 'mui-file-input'
import { useEffect, useRef, useState } from 'react'
import { useSelector } from 'react-redux'
import { useNavigate } from 'react-router-dom'
import { toast } from 'react-toastify'
import { LoadingSpinner } from '../../components/LoadingSpinner'
import { UserComponent } from '../../components/username'
import { MetadataController, NostrController } from '../../controllers'
import { appPrivateRoutes } from '../../routes'
import { State } from '../../store/rootReducer'
import { Meta, ProfileMetadata, User, UserRole } from '../../types'
import {
  encryptArrayBuffer,
  generateEncryptionKey,
  getHash,
  hexToNpub,
  npubToHex,
  queryNip05,
  sendDM,
  shorten,
  signEventForMetaFile,
  uploadToFileStorage
} from '../../utils'
import styles from './style.module.scss'
import { DndProvider } from 'react-dnd'
import { HTML5Backend } from 'react-dnd-html5-backend'
import type { Identifier, XYCoord } from 'dnd-core'
import { useDrag, useDrop } from 'react-dnd'

export const CreatePage = () => {
  const navigate = useNavigate()
  const [isLoading, setIsLoading] = useState(false)
  const [loadingSpinnerDesc, setLoadingSpinnerDesc] = useState('')

  const [authUrl, setAuthUrl] = useState<string>()

  const [selectedFiles, setSelectedFiles] = useState<File[]>([])
  const [displayUserInput, setDisplayUserInput] = useState(false)

  const [userInput, setUserInput] = useState('')
  const [userRole, setUserRole] = useState<UserRole>(UserRole.signer)
  const [error, setError] = useState<string>()

  const [users, setUsers] = useState<User[]>([])

  const usersPubkey = useSelector((state: State) => state.auth.usersPubkey)

  const nostrController = NostrController.getInstance()

  useEffect(() => {
    if (usersPubkey) {
      setUsers((prev) => {
        const existingUserIndex = prev.findIndex(
          (user) => user.pubkey === usersPubkey
        )

        // make logged in user the first signer by default
        if (existingUserIndex === -1)
          return [{ pubkey: usersPubkey, role: UserRole.signer }, ...prev]

        return prev
      })
    }
  }, [usersPubkey])

  const handleAddUser = async () => {
    setError(undefined)

    const addUser = (pubkey: string) => {
      setUsers((prev) => {
        const signers = prev.filter((user) => user.role === UserRole.signer)
        const viewers = prev.filter((user) => user.role === UserRole.viewer)

        const existingUserIndex = prev.findIndex(
          (user) => user.pubkey === pubkey
        )

        // add new
        if (existingUserIndex === -1) {
          if (userRole === UserRole.signer) {
            return [...signers, { pubkey, role: userRole }, ...viewers]
          } else {
            return [...signers, ...viewers, { pubkey, role: userRole }]
          }
        }

        const existingUser = prev[existingUserIndex]

        // return existing
        if (existingUser.role === userRole) return prev

        // change user role
        const updatedUsers = [...prev]
        const updatedUser = { ...updatedUsers[existingUserIndex] }
        updatedUser.role = userRole
        updatedUsers[existingUserIndex] = updatedUser

        // signers should be placed at the start of the  array
        return [
          ...updatedUsers.filter((user) => user.role === UserRole.signer),
          ...updatedUsers.filter((user) => user.role === UserRole.viewer)
        ]
      })
    }

    if (userInput.startsWith('npub')) {
      const pubkey = npubToHex(userInput)
      if (pubkey) {
        addUser(pubkey)
        setUserInput('')
      } else {
        setError('Provided npub is not valid. Please enter correct npub.')
      }
      return
    }

    if (userInput.includes('@')) {
      setIsLoading(true)
      setLoadingSpinnerDesc('Querying for nip05')
      const nip05Profile = await queryNip05(userInput)
        .catch((err) => {
          console.error(`error occurred in querying nip05: ${userInput}`, err)
          return null
        })
        .finally(() => {
          setIsLoading(false)
          setLoadingSpinnerDesc('')
        })

      if (nip05Profile && nip05Profile.pubkey) {
        const pubkey = nip05Profile.pubkey
        addUser(pubkey)
        setUserInput('')
      } else {
        setError('Provided nip05 is not valid. Please enter correct nip05.')
      }
      return
    }

    setError('Invalid input! Make sure to provide correct npub or nip05.')
  }

  const handleUserRoleChange = (role: UserRole, pubkey: string) => {
    setUsers((prevUsers) =>
      prevUsers.map((user) => {
        if (user.pubkey === pubkey) {
          return {
            ...user,
            role
          }
        }

        return user
      })
    )
  }

  const handleRemoveUser = (pubkey: string) => {
    setUsers((prev) => prev.filter((user) => user.pubkey !== pubkey))
  }

  /**
   * changes the position of signer in the signers list
   *
   * @param dragIndex represents the current position of user
   * @param hoverIndex represents the target position of user
   */
  const moveSigner = (dragIndex: number, hoverIndex: number) => {
    setUsers((prevUsers) => {
      const updatedUsers = [...prevUsers]
      const [draggedUser] = updatedUsers.splice(dragIndex, 1)
      updatedUsers.splice(hoverIndex, 0, draggedUser)
      return updatedUsers
    })
  }

  const handleSelectFiles = (files: File[]) => {
    setDisplayUserInput(true)
    setSelectedFiles((prev) => {
      const prevFileNames = prev.map((file) => file.name)

      const newFiles = files.filter(
        (file) => !prevFileNames.includes(file.name)
      )

      return [...prev, ...newFiles]
    })
  }

  const handleRemoveFile = (fileToRemove: File) => {
    setSelectedFiles((prevFiles) =>
      prevFiles.filter((file) => file.name !== fileToRemove.name)
    )
  }

  const handleCreate = async () => {
    if (users.length === 0) {
      toast.error(
        'No signer/viewer is provided.  At least add one signer or viewer.'
      )
      return
    }

    if (selectedFiles.length === 0) {
      toast.error('No file is selected. Select at least 1 file')
      return
    }

    setIsLoading(true)
    setLoadingSpinnerDesc('Generating hashes for files')

    const fileHashes: { [key: string]: string } = {}

    // generating file hashes
    for (const file of selectedFiles) {
      const arraybuffer = await file.arrayBuffer().catch((err) => {
        console.log(
          `err while getting arrayBuffer of file ${file.name} :>> `,
          err
        )
        toast.error(
          err.message || `err while getting arrayBuffer of file ${file.name}`
        )
        return null
      })

      if (!arraybuffer) return

      const hash = await getHash(arraybuffer)

      if (!hash) {
        setIsLoading(false)
        return
      }

      fileHashes[file.name] = hash
    }

    const zip = new JSZip()

    // zipping files
    selectedFiles.forEach((file) => {
      zip.file(`files/${file.name}`, file)
    })

    const signers = users.filter((user) => user.role === UserRole.signer)
    const viewers = users.filter((user) => user.role === UserRole.viewer)

    setLoadingSpinnerDesc('Signing nostr event')
    const createSignature = await signEventForMetaFile(
      JSON.stringify({
        signers: signers.map((signer) => hexToNpub(signer.pubkey)),
        viewers: viewers.map((viewer) => hexToNpub(viewer.pubkey)),
        fileHashes
      }),
      nostrController,
      setIsLoading
    )

    if (!createSignature) return

    // create content for meta file
    const meta: Meta = {
      createSignature: JSON.stringify(createSignature, null, 2),
      docSignatures: {}
    }

    try {
      const stringifiedMeta = JSON.stringify(meta, null, 2)
      zip.file('meta.json', stringifiedMeta)

      const metaHash = await getHash(stringifiedMeta)
      if (!metaHash) return

      const metaHashJson = {
        [usersPubkey!]: metaHash
      }

      zip.file('hashes.json', JSON.stringify(metaHashJson, null, 2))
    } catch (err) {
      console.error(err)
      toast.error('An error occurred in converting meta json to string')
      return
    }

    setLoadingSpinnerDesc('Generating zip file')

    const arraybuffer = await zip
      .generateAsync({
        type: 'arraybuffer',
        compression: 'DEFLATE',
        compressionOptions: {
          level: 6
        }
      })
      .catch((err) => {
        console.log('err in zip:>> ', err)
        setIsLoading(false)
        toast.error(err.message || 'Error occurred in generating zip file')
        return null
      })

    if (!arraybuffer) return

    const encryptionKey = await generateEncryptionKey()

    setLoadingSpinnerDesc('Encrypting zip file')
    const encryptedArrayBuffer = await encryptArrayBuffer(
      arraybuffer,
      encryptionKey
    )

    const blob = new Blob([encryptedArrayBuffer])

    setLoadingSpinnerDesc('Uploading zip file to file storage.')
    const fileUrl = await uploadToFileStorage(blob, nostrController)
      .then((url) => {
        toast.success('zip file uploaded to file storage')
        return url
      })
      .catch((err) => {
        console.log('err in upload:>> ', err)
        setIsLoading(false)
        toast.error(err.message || 'Error occurred in uploading zip file')
        return null
      })

    if (!fileUrl) return

    setLoadingSpinnerDesc('Sending DM to signers/viewers')

    // send DM to first signer if exists
    if (signers.length > 0) {
      await sendDM(
        fileUrl,
        encryptionKey,
        signers[0].pubkey,
        nostrController,
        true,
        setAuthUrl
      )
    } else {
      // send DM to all viewers if no signer
      for (const viewer of viewers) {
        // todo: execute in parallel
        await sendDM(
          fileUrl,
          encryptionKey,
          viewer.pubkey,
          nostrController,
          false,
          setAuthUrl
        )
      }
    }
    setIsLoading(false)

    navigate(
      `${appPrivateRoutes.sign}?file=${encodeURIComponent(
        fileUrl
      )}&key=${encodeURIComponent(encryptionKey)}`
    )
  }

  if (authUrl) {
    return (
      <iframe
        title="Nsecbunker auth"
        src={authUrl}
        width="100%"
        height="500px"
      />
    )
  }

  return (
    <>
      {isLoading && <LoadingSpinner desc={loadingSpinnerDesc} />}
      <Box className={styles.container}>
        <Typography component="label" variant="h6">
          Select files
        </Typography>

        <MuiFileInput
          multiple
          placeholder="Choose Files"
          value={selectedFiles}
          onChange={(value) => handleSelectFiles(value)}
        />

        <ul>
          {selectedFiles.map((file, index) => (
            <li key={index}>
              <Typography component="label">{file.name}</Typography>
              <IconButton onClick={() => handleRemoveFile(file)}>
                <Clear style={{ color: 'red' }} />{' '}
              </IconButton>
            </li>
          ))}
        </ul>

        {displayUserInput && (
          <>
            <Typography component="label" variant="h6">
              Add Counterparties
            </Typography>
            <Box className={styles.inputBlock}>
              <TextField
                label="nip05 / npub"
                value={userInput}
                onChange={(e) => setUserInput(e.target.value)}
                helperText={error}
                error={!!error}
              />
              <FormControl fullWidth>
                <InputLabel id="select-role-label">Role</InputLabel>
                <Select
                  labelId="select-role-label"
                  id="demo-simple-select"
                  value={userRole}
                  label="Role"
                  onChange={(e) => setUserRole(e.target.value as UserRole)}
                >
                  <MenuItem value={UserRole.signer}>{UserRole.signer}</MenuItem>
                  <MenuItem value={UserRole.viewer}>{UserRole.viewer}</MenuItem>
                </Select>
              </FormControl>

              <Box sx={{ mt: 1, display: 'flex', justifyContent: 'center' }}>
                <Button
                  disabled={!userInput}
                  onClick={handleAddUser}
                  variant="contained"
                >
                  Add
                </Button>
              </Box>
            </Box>
            <DisplayUser
              users={users}
              handleUserRoleChange={handleUserRoleChange}
              handleRemoveUser={handleRemoveUser}
              moveSigner={moveSigner}
            />
            <Box sx={{ mt: 1, display: 'flex', justifyContent: 'center' }}>
              <Button onClick={handleCreate} variant="contained">
                Create
              </Button>
            </Box>
          </>
        )}
      </Box>
    </>
  )
}

type DisplayUsersProps = {
  users: User[]
  handleUserRoleChange: (role: UserRole, pubkey: string) => void
  handleRemoveUser: (pubkey: string) => void
  moveSigner: (dragIndex: number, hoverIndex: number) => void
}

const DisplayUser = ({
  users,
  handleUserRoleChange,
  handleRemoveUser,
  moveSigner
}: DisplayUsersProps) => {
  const [metadata, setMetadata] = useState<{ [key: string]: ProfileMetadata }>(
    {}
  )

  useEffect(() => {
    users.forEach((user) => {
      if (!(user.pubkey in metadata)) {
        const metadataController = new MetadataController()

        metadataController
          .findMetadata(user.pubkey)
          .then((metadataEvent) => {
            const metadataContent =
              metadataController.extractProfileMetadataContent(metadataEvent)
            if (metadataContent)
              setMetadata((prev) => ({
                ...prev,
                [user.pubkey]: metadataContent
              }))
          })
          .catch((err) => {
            console.error(
              `error occurred in finding metadata for: ${user.pubkey}`,
              err
            )
          })
      }
    })
  }, [users])

  return (
    <TableContainer component={Paper} elevation={3} sx={{ marginTop: '20px' }}>
      <Table>
        <TableHead>
          <TableRow>
            <TableCell className={styles.tableHeaderCell}>User</TableCell>
            <TableCell className={styles.tableHeaderCell}>Role</TableCell>
            <TableCell>Action</TableCell>
          </TableRow>
        </TableHead>
        <TableBody>
          <DndProvider backend={HTML5Backend}>
            {users
              .filter((user) => user.role === UserRole.signer)
              .map((user, index) => (
                <SignerRow
                  key={`signer-${index}`}
                  userMeta={metadata[user.pubkey]}
                  user={user}
                  index={index}
                  moveSigner={moveSigner}
                  handleUserRoleChange={handleUserRoleChange}
                  handleRemoveUser={handleRemoveUser}
                />
              ))}
          </DndProvider>
          {users
            .filter((user) => user.role === UserRole.viewer)
            .map((user, index) => {
              const userMeta = metadata[user.pubkey]
              return (
                <TableRow key={index}>
                  <TableCell className={styles.tableCell}>
                    <UserComponent
                      pubkey={user.pubkey}
                      name={
                        userMeta?.display_name ||
                        userMeta?.name ||
                        shorten(hexToNpub(user.pubkey))
                      }
                      image={userMeta?.picture}
                    />
                  </TableCell>
                  <TableCell className={styles.tableCell}>
                    <Select
                      fullWidth
                      value={user.role}
                      onChange={(e) =>
                        handleUserRoleChange(
                          e.target.value as UserRole,
                          user.pubkey
                        )
                      }
                    >
                      <MenuItem value={UserRole.signer}>
                        {UserRole.signer}
                      </MenuItem>
                      <MenuItem value={UserRole.viewer}>
                        {UserRole.viewer}
                      </MenuItem>
                    </Select>
                  </TableCell>
                  <TableCell>
                    <Tooltip title="Remove User" arrow>
                      <IconButton onClick={() => handleRemoveUser(user.pubkey)}>
                        <Clear style={{ color: 'red' }} />
                      </IconButton>
                    </Tooltip>
                  </TableCell>
                </TableRow>
              )
            })}
        </TableBody>
      </Table>
    </TableContainer>
  )
}

interface DragItem {
  index: number
  id: string
  type: string
}

type SignerRowProps = {
  userMeta: ProfileMetadata
  user: User
  index: number
  moveSigner: (dragIndex: number, hoverIndex: number) => void
  handleUserRoleChange: (role: UserRole, pubkey: string) => void
  handleRemoveUser: (pubkey: string) => void
}

const SignerRow = ({
  userMeta,
  user,
  index,
  moveSigner,
  handleUserRoleChange,
  handleRemoveUser
}: SignerRowProps) => {
  const ref = useRef<HTMLTableRowElement>(null)

  const [{ handlerId }, drop] = useDrop<
    DragItem,
    void,
    { handlerId: Identifier | null }
  >({
    accept: 'row',
    collect(monitor) {
      return {
        handlerId: monitor.getHandlerId()
      }
    },
    hover(item: DragItem, monitor) {
      if (!ref.current) {
        return
      }
      const dragIndex = item.index
      const hoverIndex = index

      // Don't replace items with themselves
      if (dragIndex === hoverIndex) {
        return
      }

      // Determine rectangle on screen
      const hoverBoundingRect = ref.current?.getBoundingClientRect()

      // Get vertical middle
      const hoverMiddleY =
        (hoverBoundingRect.bottom - hoverBoundingRect.top) / 2

      // Determine mouse position
      const clientOffset = monitor.getClientOffset()

      // Get pixels to the top
      const hoverClientY = (clientOffset as XYCoord).y - hoverBoundingRect.top

      // Only perform the move when the mouse has crossed half of the items height
      // When dragging downwards, only move when the cursor is below 50%
      // When dragging upwards, only move when the cursor is above 50%

      // Dragging downwards
      if (dragIndex < hoverIndex && hoverClientY < hoverMiddleY) {
        return
      }

      // Dragging upwards
      if (dragIndex > hoverIndex && hoverClientY > hoverMiddleY) {
        return
      }

      // Time to actually perform the action
      moveSigner(dragIndex, hoverIndex)

      // Note: we're mutating the monitor item here!
      // Generally it's better to avoid mutations,
      // but it's good here for the sake of performance
      // to avoid expensive index searches.
      item.index = hoverIndex
    }
  })

  const [{ isDragging }, drag] = useDrag({
    type: 'row',
    item: () => {
      return { id: user.pubkey, index }
    },
    collect: (monitor: any) => ({
      isDragging: monitor.isDragging()
    })
  })

  const opacity = isDragging ? 0 : 1
  drag(drop(ref))

  return (
    <TableRow
      sx={{ cursor: 'move', opacity }}
      data-handler-id={handlerId}
      ref={ref}
    >
      <TableCell
        className={styles.tableCell}
        sx={{ display: 'flex', alignItems: 'center', gap: '10px' }}
      >
        <DragHandle />
        <UserComponent
          pubkey={user.pubkey}
          name={
            userMeta?.display_name ||
            userMeta?.name ||
            shorten(hexToNpub(user.pubkey))
          }
          image={userMeta?.picture}
        />
      </TableCell>
      <TableCell className={styles.tableCell}>
        <Select
          fullWidth
          value={user.role}
          onChange={(e) =>
            handleUserRoleChange(e.target.value as UserRole, user.pubkey)
          }
        >
          <MenuItem value={UserRole.signer}>{UserRole.signer}</MenuItem>
          <MenuItem value={UserRole.viewer}>{UserRole.viewer}</MenuItem>
        </Select>
      </TableCell>
      <TableCell>
        <Tooltip title="Remove User" arrow>
          <IconButton onClick={() => handleRemoveUser(user.pubkey)}>
            <Clear style={{ color: 'red' }} />
          </IconButton>
        </Tooltip>
      </TableCell>
    </TableRow>
  )
}
